#!/usr/bin/env python
from setuptools import setup, find_packages

setup(
    name='cfw',
    version='0.1.11',
    url='https://github.com/zinic/python-cfw',
    description='Python Command Line Interface Library',

    author='John Hopper',
    author_email='john.hopper@jpserver.net',

    packages=find_packages(),

    license='MIT'
)
